;;; finder-inf.el --- keyword-to-package mapping
;; Keywords: help
;;; Commentary:
;; Don't edit this file.  It's generated by finder.el

;;; Code:

(setq finder-package-info '(
    ("abbrev.el"
        "abbrev mode commands for Emacs"
        (abbrev))
    ("abbrevlist.el"
        "list one abbrev table alphabetically ordered."
        (abbrev))
    ("ada-mode.el"
        "An Emacs major-mode for editing Ada source."
        nil)
    ("add-log.el"
        "change log maintenance commands for Emacs"
        (maint))
    ("advice.el"
        "an overloading mechanism for Emacs Lisp functions"
        (extensions  lisp  tools))
    ("allout.el"
        "Extensive outline mode for use alone and with other modes."
        (outline mode))
    ("ange-ftp.el"
        "transparent FTP support for GNU Emacs"
        (comm))
    ("appt.el"
        "appointment notification functions."
        (calendar))
    ("apropos.el"
        "apropos commands for users and programmers."
        (help))
    ("arc-mode.el"
        "simple editing of archives"
        (archives msdog editing major-mode))
    ("array.el"
        "array editing commands for Gnu Emacs"
        (extensions))
    ("asm-mode.el"
        "mode for editing assembler code"
        (tools  languages))
    ("assoc.el"
        "insert/delete/sort functions on association lists"
        (extensions))
    ("auto-show.el"
        "perform automatic horizontal scrolling as point moves"
        (scroll display minor-mode))
    ("autoinsert.el"
        "automatic mode-dependent insertion of text into new files"
        nil)
    ("autoload.el"
        "maintain autoloads in loaddefs.el."
        (maint))
    ("avoid.el"
        "make mouse pointer stay out of the way of editing"
        (mouse))
    ("awk-mode.el"
        "AWK code editing commands for Emacs"
        (unix  languages))
    ("backquote.el"
        "implement the ` Lisp construct"
        (extensions  internal))
    ("bib-mode.el"
        "bib-mode, major mode for editing bib files."
        (bib))
    ("bibtex.el"
        "BibTeX mode for GNU Emacs"
        (bibtex  latex  tex))
    ("blackbox.el"
        "blackbox game in Emacs Lisp"
        (games))
    ("blessmail.el"
        "Decide whether movemail needs special privileges."
        (internal))
    ("bookmark.el"
        "set bookmarks, maybe annotate them, jump to them later."
        (bookmarks  placeholders  annotations))
    ("browse-url.el"
        "ask a WWW browser to load a URL"
        (hypertext))
    ("buff-menu.el"
        "buffer menu main function and support functions."
        nil)
    ("byte-opt.el"
        "the optimization passes of the emacs-lisp byte compiler."
        (internal))
    ("byte-run.el"
        "byte-compiler support for inlining"
        (internal))
    ("bytecomp.el"
        "compilation of Lisp code into byte code."
        (internal))
    ("c-mode.el"
        "C code editing commands for Emacs"
        (c))
    ("cal-china.el"
        "calendar functions for the Chinese calendar."
        (calendar))
    ("cal-coptic.el"
        "calendar functions for the Coptic/Ethiopic calendars."
        (calendar))
    ("cal-dst.el"
        "calendar functions for daylight savings rules."
        (calendar))
    ("cal-french.el"
        "calendar functions for the French Revolutionary calendar."
        (calendar))
    ("cal-hebrew.el"
        "calendar functions for the Hebrew calendar."
        (calendar))
    ("cal-islam.el"
        "calendar functions for the Islamic calendar."
        (calendar))
    ("cal-iso.el"
        "calendar functions for the ISO calendar."
        (calendar))
    ("cal-julian.el"
        "calendar functions for the Julian calendar."
        (calendar))
    ("cal-mayan.el"
        "calendar functions for the Mayan calendars."
        (calendar))
    ("cal-menu.el"
        "calendar functions for menu bar and popup menu support"
        (calendar))
    ("cal-move.el"
        "calendar functions for movement in the calendar"
        (calendar))
    ("cal-persia.el"
        "calendar functions for the Persian calendar."
        (calendar))
    ("cal-tex.el"
        "calendar functions for printing calendars with LaTeX."
        (calendar))
    ("cal-x.el"
        "calendar windows in dedicated frames in X"
        (calendar))
    ("calendar.el"
        "Calendar functions."
        (calendar))
    ("case-table.el"
        "code to extend the character set and support case tables."
        (i18n))
    ("cc-compat.el"
        "cc-mode compatibility with c-mode.el confusion"
        (c languages oop))
    ("cc-mode.el"
        "major mode for editing C, C++, and Objective-C code"
        (c languages oop))
    ("cdl.el"
        "Common Data Language (CDL) utility functions for Gnu Emacs"
        (data))
    ("chistory.el"
        "list command history"
        nil)
    ("cl-compat.el"
        "Common Lisp extensions for GNU Emacs Lisp (compatibility)"
        (extensions))
    ("cl-extra.el"
        "Common Lisp extensions for GNU Emacs Lisp (part two)"
        (extensions))
    ("cl-indent.el"
        "enhanced lisp-indent mode"
        (lisp  tools))
    ("cl-macs.el"
        "Common Lisp extensions for GNU Emacs Lisp (part four)"
        (extensions))
    ("cl-seq.el"
        "Common Lisp extensions for GNU Emacs Lisp (part three)"
        (extensions))
    ("cl-specs.el"
        "Edebug specs for cl.el"
        (lisp  tools  maint))
    ("cl.el"
        "Common Lisp extensions for GNU Emacs Lisp"
        (extensions))
    ("cmacexp.el"
        "expand C macros in a region"
        (c))
    ("cmuscheme.el"
        "Scheme process in a buffer. Adapted from tea.el."
        (processes  lisp))
    ("comint.el"
        "general command interpreter in a window stuff"
        (processes))
    ("compare-w.el"
        "compare text between windows for Emacs."
        nil)
    ("compile.el"
        "run compiler as inferior of Emacs, parse error messages."
        (tools  processes))
    ("complete.el"
        "partial completion mechanism plus other goodies"
        (abbrev))
    ("completion.el"
        "dynamic word-completion code"
        (abbrev))
    ("cookie1.el"
        "retrieve random phrases from fortune cookie files"
        (games))
    ("copyright.el"
        "update the copyright notice in current buffer"
        (maint  tools))
    ("cplus-md.el"
        "C++ code editing commands for Emacs"
        (c))
    ("cpp.el"
        "Highlight or hide text according to cpp conditionals."
        (c  faces  tools))
    ("cust-print.el"
        "handles print-level and print-circle."
        (extensions))
    ("custom.el"
        "User friendly customization support."
        (help))
    ("dabbrev.el"
        "dynamic abbreviation package"
        (abbrev expand completion))
    ("debug.el"
        "debuggers and related commands for Emacs"
        (lisp  tools  maint))
    ("decipher.el"
        "Cryptanalyze monoalphabetic substitution ciphers"
        (games))
    ("default.el"
        nil
        nil)
    ("delsel.el"
        "delete selection if you insert"
        nil)
    ("derived.el"
        "allow inheritance of major modes."
        nil)
    ("desktop.el"
        "save partial status of Emacs when killed"
        (customization))
    ("diary-lib.el"
        "diary functions."
        (calendar))
    ("diff.el"
        "Run `diff' in compilation-mode."
        (unix  tools))
    ("dired-aux.el"
        "less commonly used parts of dired  -*-byte-compile-dynamic: t;-*-"
        nil)
    ("dired-x.el"
        "Sebastian Kremer's Extra DIRED hacked up for GNU Emacs19"
        (dired extensions))
    ("dired.el"
        "directory-browsing commands"
        nil)
    ("disass.el"
        "disassembler for compiled Emacs Lisp code"
        (internal))
    ("disp-table.el"
        "functions for dealing with char tables."
        (i18n))
    ("dissociate.el"
        "scramble text amusingly for Emacs."
        (games))
    ("docref.el"
        "Simple cross references for Elisp documentation strings"
        (docs  help  lisp))
    ("doctor.el"
        "psychological help for frustrated users."
        (games))
    ("dos-fns.el"
        "MS-Dos specific functions."
        (internal))
    ("double.el"
        "Support for keyboard remapping with double clicking"
        (i18n))
    ("dunnet.el"
        "Text adventure for Emacs"
        (games))
    ("easymenu.el"
        "support the easymenu interface for defining a menu."
        (emulations))
    ("ebuff-menu.el"
        "electric-buffer-list mode"
        nil)
    ("echistory.el"
        "Electric Command History Mode"
        nil)
    ("edebug.el"
        "a source-level debugger for Emacs Lisp"
        (lisp  tools  maint))
    ("ediff-diff.el"
        "diff-related utilities"
        nil)
    ("ediff-hook.el"
        "setup for Ediff's menus and autoloads"
        nil)
    ("ediff-init.el"
        "Macros, variables, and defsubsts used by Ediff"
        nil)
    ("ediff-merg.el"
        "merging utilities"
        nil)
    ("ediff-mult.el"
        "support for multi-file/multi-buffer processing in Ediff"
        nil)
    ("ediff-ptch.el"
        "Ediff's  patch support"
        nil)
    ("ediff-util.el"
        "the core commands and utilities of ediff"
        nil)
    ("ediff-vers.el"
        "version control interface to Ediff"
        nil)
    ("ediff-wind.el"
        "window manipulation utilities"
        nil)
    ("ediff.el"
        "a comprehensive visual interface to diff & patch"
        (comparing  merging  patching  version control.))
    ("edmacro.el"
        "keyboard macro editor"
        (abbrev))
    ("edt-lk201.el"
        "Enhanced EDT Keypad Mode Emulation for LK-201 Keyboards"
        (emulations))
    ("edt-mapper.el"
        "Create an EDT LK-201 Map File for X-Windows Emacs"
        (emulations))
    ("edt-pc.el"
        "Enhanced EDT Keypad Mode Emulation for PC 101 Keyboards"
        (emulations))
    ("edt-vt100.el"
        "Enhanced EDT Keypad Mode Emulation for VT Series Terminals"
        (emulations))
    ("edt.el"
        "Enhanced EDT Keypad Mode Emulation for GNU Emacs 19"
        (emulations))
    ("ehelp.el"
        "bindings for electric-help mode"
        (help  extensions))
    ("electric.el"
        "window maker and Command loop for `electric' modes."
        (extensions))
    ("elp.el"
        "Emacs Lisp Profiler"
        (emacs lisp profile timing))
    ("emacs-lock.el"
        "prevents you from exiting emacs if a buffer is locked"
        nil)
    ("emacsbug.el"
        "command to report Emacs bugs to appropriate mailing list."
        (maint))
    ("emerge.el"
        "merge diffs under Emacs control"
        (unix  tools))
    ("enriched.el"
        "read and save files in text/enriched format"
        (wp  faces))
    ("env.el"
        "functions to manipulate environment variables."
        (processes  unix))
    ("etags.el"
        "etags facility for Emacs"
        (tools))
    ("eval-reg.el"
        "Redefine eval-region, and subrs that use it, in Lisp"
        (lisp))
    ("executable.el"
        "base functionality for executable interpreter scripts"
        (languages  unix))
    ("f90.el"
        "Fortran-90 mode (free format)"
        (fortran  f90  languages))
    ("facemenu.el"
        "create a face menu for interactively adding fonts to text"
        (faces))
    ("faces.el"
        "Lisp interface to the c \"face\" structure"
        nil)
    ("fast-lock.el"
        "Automagic text properties caching for fast Font Lock mode."
        (faces files))
    ("ffap.el"
        "find file or url at point"
        (files  hypermedia  matching  mouse))
    ("files.el"
        "file input and output commands for Emacs"
        nil)
    ("fill.el"
        "fill commands for Emacs"
        (wp))
    ("find-dired.el"
        "run a `find' command and dired the output"
        (unix))
    ("find-file.el"
        "find a file corresponding to this one given a pattern"
        (c  matching  tools))
    ("find-gc.el"
        "detect functions that call the garbage collector"
        nil)
    ("finder-inf.el"
        "keyword-to-package mapping"
        (help))
    ("finder.el"
        "topic & keyword-based code finder"
        (help))
    ("float-sup.el"
        "detect absence of floating-point support in Emacs runtime"
        nil)
    ("float.el"
        "floating point arithmetic package."
        (extensions))
    ("flow-ctrl.el"
        "help for lusers on cu(1) or ttys with wired-in ^S/^Q flow control"
        (hardware))
    ("foldout.el"
        "Folding extensions for outline-mode and outline-minor-mode."
        (folding  outline))
    ("follow.el"
        "Minor mode, Synchronize windows showing the same buffer."
        (display  window  minor-mode))
    ("font-lock.el"
        "Electric font lock mode"
        (languages  faces))
    ("format.el"
        "read and save files in multiple formats"
        nil)
    ("forms-d2.el"
        "demo forms-mode"
        nil)
    ("forms-pass.el"
        "passwd file demo for forms-mode"
        nil)
    ("forms.el"
        "Forms mode: edit a file as a form to fill in"
        nil)
    ("fortran.el"
        "Fortran mode for GNU Emacs"
        (languages))
    ("frame.el"
        "multi-frame management independent of window systems."
        (internal))
    ("gnus-cache.el"
        "cache interface for Gnus"
        (news))
    ("gnus-cite.el"
        "parse citations in articles for Gnus"
        (news  mail))
    ("gnus-cus.el"
        "User friendly customization of Gnus"
        (help  news))
    ("gnus-demon.el"
        "daemonic Gnus behaviour"
        (news))
    ("gnus-edit.el"
        "Gnus SCORE file editing"
        (news  help))
    ("gnus-ems.el"
        "functions for making Gnus work under different Emacsen"
        (news))
    ("gnus-gl.el"
        "an interface to GroupLens for Gnus"
        (news  score))
    ("gnus-kill.el"
        "kill commands for Gnus"
        (news))
    ("gnus-mh.el"
        "mh-e interface for Gnus"
        (news))
    ("gnus-msg.el"
        "mail and post interface for Gnus"
        (news))
    ("gnus-nocem.el"
        "NoCeM pseudo-cancellation treatment"
        (news))
    ("gnus-salt.el"
        "alternate summary mode interfaces for Gnus"
        nil)
    ("gnus-score.el"
        "scoring code for Gnus"
        (news))
    ("gnus-setup.el"
        "Initialization & Setup for Gnus 5"
        (news))
    ("gnus-soup.el"
        "SOUP packet writing support for Gnus"
        (news  mail))
    ("gnus-srvr.el"
        "virtual server support for Gnus"
        (news))
    ("gnus-topic.el"
        "a folding minor mode for Gnus group buffers"
        (news))
    ("gnus-uu.el"
        "extract (uu)encoded files in Gnus"
        nil)
    ("gnus-vis.el"
        "display-oriented parts of Gnus"
        (news))
    ("gnus-vm.el"
        "vm interface for Gnus"
        (news  mail))
    ("gnus.el"
        "a newsreader for GNU Emacs"
        (news))
    ("gomoku.el"
        "Gomoku game between you and Emacs"
        (games))
    ("goto-addr.el"
        "click to browse URL or to send to e-mail address"
        (mh-e  www  mouse  mail))
    ("gud.el"
        "Grand Unified Debugger mode for gdb, sdb, dbx, or xdb under Emacs"
        (unix  tools))
    ("gulp.el"
        "Ask for updates for Lisp packages"
        (maintenance))
    ("hanoi.el"
        "towers of hanoi in GNUmacs"
        (games))
    ("help-macro.el"
        "Makes command line help such as help-for-help"
        nil)
    ("help.el"
        "help commands for Emacs"
        (help  internal))
    ("helper.el"
        "utility help package supporting help in electric modes"
        (help))
    ("hexl.el"
        "edit a file in a hex dump format using the hexl filter."
        (data))
    ("hideif.el"
        "hides selected code within ifdef."
        (c  outlines))
    ("hideshow.el"
        "minor mode cmds to selectively display blocks of code"
        (c c++ lisp tools editing))
    ("hilit19.el"
        "customizable highlighting for Emacs19"
        (faces))
    ("hippie-exp.el"
        "expand text trying various ways to find its expansion."
        (abbrev))
    ("holidays.el"
        "holiday functions for the calendar package"
        (holidays  calendar))
    ("icomplete.el"
        "minibuffer completion incremental feedback"
        (help  abbrev))
    ("icon.el"
        "mode for editing Icon code"
        (languages))
    ("ielm.el"
        "interaction mode for Emacs Lisp"
        (lisp))
    ("imenu.el"
        "Framework for mode-specific buffer indexes."
        (tools))
    ("indent.el"
        "indentation commands for Emacs"
        nil)
    ("inf-lisp.el"
        "an inferior-lisp mode"
        (processes  lisp))
    ("info.el"
        "info package for Emacs."
        (help))
    ("informat.el"
        "info support functions package for Emacs"
        (help))
    ("isearch.el"
        "incremental search minor mode."
        nil)
    ("iso-acc.el"
        "minor mode providing electric accent keys"
        (i18n))
    ("iso-ascii.el"
        "set up char tables for ISO 8859/1 on ASCII terminals."
        (i18n))
    ("iso-cvt.el"
        "translate to ISO 8859-1 from/to net/TeX conventions"
        (tex  iso  latin  i18n))
    ("iso-insert.el"
        "insert functions for ISO 8859/1."
        (i18n))
    ("iso-swed.el"
        "set up char tables for ISO 8859/1 for Swedish/Finnish ttys"
        (i18n))
    ("iso-syntax.el"
        "set up case-conversion and syntax tables for ISO 8859/1"
        (i18n))
    ("iso-transl.el"
        "keyboard input definitions for ISO 8859/1."
        (i18n))
    ("iso02-syn.el"
        "set up case-conversion and syntax tables for ISO 8859-2"
        (i18n))
    ("ispell.el"
        "spell checking using Ispell"
        nil)
    ("ispell4.el"
        "this is the GNU EMACS interface to GNU ISPELL version 4."
        (wp))
    ("jka-compr.el"
        "reading/writing/loading compressed files"
        (data))
    ("kermit.el"
        "additions to shell mode for use with kermit, etc."
        (comm))
    ("lazy-lock.el"
        "Lazy demand-driven fontification for fast Font Lock mode."
        (faces files))
    ("ledit.el"
        "Emacs side of ledit interface"
        nil)
    ("levents.el"
        "emulate the Lucid event data type and associated functions."
        nil)
    ("life.el"
        "John Horton Conway's `Life' game for GNU Emacs"
        (games))
    ("lisp-mnt.el"
        "minor mode for Emacs Lisp maintainers"
        (docs))
    ("lisp-mode.el"
        "Lisp mode, and its idiosyncratic commands."
        (lisp  languages))
    ("lisp.el"
        "Lisp editing commands for Emacs"
        (lisp  languages))
    ("lmenu.el"
        "emulate Lucid's menubar support"
        (emulations))
    ("loaddefs.el"
        "define standard autoloads and keys of other files"
        (internal))
    ("loadhist.el"
        "lisp functions for working with feature groups"
        (internal))
    ("loadup.el"
        "load up standardly loaded Lisp files for Emacs."
        (internal))
    ("lpr.el"
        "print Emacs buffer on line printer."
        (unix))
    ("ls-lisp.el"
        "emulate insert-directory completely in Emacs Lisp"
        (unix))
    ("lselect.el"
        "Lucid interface to X Selections"
        (emulations))
    ("lucid.el"
        "Emulate some Lucid Emacs functions."
        nil)
    ("lunar.el"
        "calendar functions for phases of the moon."
        (calendar))
    ("macros.el"
        "non-primitive commands for keyboard macros."
        (abbrev))
    ("mail-extr.el"
        "extract full name and address from RFC 822 mail header."
        (mail))
    ("mail-hist.el"
        "Headers and message body history for outgoing mail."
        (mail  history))
    ("mail-utils.el"
        "utility functions used both by rmail and rnews"
        (mail  news))
    ("mailabbrev.el"
        "abbrev-expansion of mail aliases."
        (mail))
    ("mailalias.el"
        "expand and complete mailing address aliases"
        (mail))
    ("mailheader.el"
        "Mail header parsing, merging, formatting"
        (tools  mail  news))
    ("mailpost.el"
        "RMAIL coupler to /usr/uci/post mailer"
        (mail))
    ("make-mode.el"
        "makefile editing commands for Emacs"
        (unix  tools))
    ("makeinfo.el"
        "run makeinfo conveniently"
        nil)
    ("makesum.el"
        "generate key binding summary for Emacs"
        (help))
    ("man.el"
        "browse UNIX manual pages"
        (help))
    ("map-ynp.el"
        "General-purpose boolean question-asker."
        (lisp  extensions))
    ("meese.el"
        "protect the impressionable young minds of America"
        (games))
    ("menu-bar.el"
        "define a default menu bar."
        (internal))
    ("message.el"
        "composing mail and news messages"
        (mail  news))
    ("metamail.el"
        "Metamail interface for GNU Emacs"
        (mail  news  mime  multimedia))
    ("mh-comp.el"
        "mh-e functions for composing messages"
        nil)
    ("mh-e.el"
        "GNU Emacs interface to the MH mail system"
        (mail))
    ("mh-funcs.el"
        "mh-e functions not everyone will use right away"
        nil)
    ("mh-mime.el"
        "mh-e support for composing MIME messages"
        nil)
    ("mh-pick.el"
        "make a search pattern and search for a message in mh-e"
        nil)
    ("mh-seq.el"
        "mh-e sequences support"
        nil)
    ("mh-utils.el"
        "mh-e code needed for both sending and reading"
        nil)
    ("misc.el"
        "basic editing commands for Emacs"
        nil)
    ("mlconvert.el"
        "convert buffer of Mocklisp code to real lisp."
        (emulations))
    ("mldrag.el"
        "mode line and vertical line dragging to resize windows"
        (mouse))
    ("mlsupport.el"
        "run-time support for mocklisp code."
        (extensions))
    ("modula2.el"
        "Modula-2 editing support package"
        (languages))
    ("morse.el"
        "Convert text to morse code and back."
        nil)
    ("mouse-sel.el"
        "Multi-click selection support for Emacs 19"
        (mouse))
    ("mouse.el"
        "window system-independent mouse support"
        (hardware))
    ("mpuz.el"
        "multiplication puzzle for GNU Emacs"
        (games))
    ("msb.el"
        "Customizable buffer-selection with multiple menus."
        (mouse buffer menu ))
    ("nnbabyl.el"
        "rmail mbox access for Gnus"
        (news  mail))
    ("nndb.el"
        "nndb access for Gnus"
        (news))
    ("nndir.el"
        "single directory newsgroup access for Gnus"
        (news))
    ("nndoc.el"
        "single file access for Gnus"
        (news))
    ("nneething.el"
        "random file access for Gnus"
        (news  mail))
    ("nnfolder.el"
        "mail folder access for Gnus"
        (mail))
    ("nnheader.el"
        "header access macros for Gnus and its backends"
        (news))
    ("nnkiboze.el"
        "select virtual news access for Gnus"
        (news))
    ("nnmail.el"
        "mail support functions for the Gnus mail backends"
        (news  mail))
    ("nnmbox.el"
        "mail mbox access for Gnus"
        (news  mail))
    ("nnmh.el"
        "mhspool access for Gnus"
        (news  mail))
    ("nnml.el"
        "mail spool access for Gnus"
        (news  mail))
    ("nnoo.el"
        "OO Gnus Backends"
        (news))
    ("nnsoup.el"
        "SOUP access for Gnus"
        (news  mail))
    ("nnspool.el"
        "spool access for GNU Emacs"
        (news))
    ("nntp.el"
        "nntp access for Gnus"
        (news))
    ("nnvirtual.el"
        "virtual newsgroups access for Gnus"
        (news))
    ("noutline.el"
        "outline mode commands for Emacs"
        (outlines))
    ("novice.el"
        "handling of disabled commands (\"novice mode\") for Emacs."
        (internal  help))
    ("nroff-mode.el"
        "GNU Emacs major mode for editing nroff source"
        (wp))
    ("options.el"
        "edit Options command for Emacs."
        nil)
    ("outline.el"
        "outline mode commands for Emacs"
        (outlines))
    ("page-ext.el"
        "extended page handling commands"
        nil)
    ("page.el"
        "page motion commands for emacs."
        nil)
    ("paragraphs.el"
        "paragraph and sentence parsing."
        (wp))
    ("paren.el"
        "highlight matching paren."
        (languages  faces))
    ("pascal.el"
        "major mode for editing pascal source in Emacs"
        (languages))
    ("patcomp.el"
        nil
        nil)
    ("paths.el"
        "define pathnames for use by various Emacs commands."
        (internal))
    ("pc-mode.el"
        "emulate certain key bindings used on PCs."
        (emulations))
    ("pc-select.el"
        "emulate mark, cut, copy and paste from motif"
        nil)
    ("perl-mode.el"
        "Perl code editing commands for GNU Emacs"
        (languages))
    ("picture.el"
        "\"Picture mode\" -- editing using quarter-plane screen model."
        nil)
    ("pp.el"
        "pretty printer for Emacs Lisp"
        nil)
    ("profile.el"
        "generate run time measurements of Emacs Lisp functions"
        (lisp  tools))
    ("prolog.el"
        "major mode for editing and running Prolog under Emacs"
        (languages))
    ("ps-print.el"
        "Jim's Pretty-Good PostScript Generator for Emacs 19."
        (print  postscript))
    ("rcompile.el"
        "run a compilation on a remote machine"
        (tools  processes))
    ("rect.el"
        "rectangle functions for GNU Emacs."
        (internal))
    ("refbib.el"
        "convert refer-style references to ones usable by Latex bib"
        (bib  tex))
    ("refer.el"
        "look up references in bibliography files."
        (bib))
    ("regi.el"
        "REGular expression Interpreting engine"
        (extensions  matching))
    ("register.el"
        "register commands for Emacs."
        (internal))
    ("replace.el"
        "replace commands for Emacs."
        nil)
    ("reporter.el"
        "customizable bug reporting of lisp programs"
        (maint mail tools))
    ("reposition.el"
        "center a Lisp function or comment on the screen"
        nil)
    ("resume.el"
        "process command line args from within a suspended Emacs job"
        (processes))
    ("rfc822.el"
        "hairy rfc822 parser for mail and news and suchlike"
        (mail))
    ("ring.el"
        "handle rings of items"
        (extensions))
    ("rlogin.el"
        "remote login interface"
        (unix  comm))
    ("rmail.el"
        "main code of \"RMAIL\" mail reader for Emacs."
        (mail))
    ("rmailedit.el"
        "\"RMAIL edit mode\"  Edit the current message."
        (mail))
    ("rmailkwd.el"
        "part of the \"RMAIL\" mail reader for Emacs."
        (mail))
    ("rmailmsc.el"
        "miscellaneous support functions for the RMAIL mail reader"
        (mail))
    ("rmailout.el"
        "\"RMAIL\" mail reader for Emacs: output message to a file."
        (mail))
    ("rmailsort.el"
        "Rmail: sort messages."
        (mail))
    ("rmailsum.el"
        "make summary buffers for the mail reader"
        (mail))
    ("rnews.el"
        "USENET news reader for gnu emacs"
        (news))
    ("rnewspost.el"
        "USENET news poster/mailer for GNU Emacs"
        (mail  news))
    ("rot13.el"
        "display a buffer in rot13."
        nil)
    ("rsz-mini.el"
        "dynamically resize minibuffer to display entire contents"
        (minibuffer  window  frame  display))
    ("s-region.el"
        "set region using shift key."
        (terminals))
    ("saveplace.el"
        "automatically save place in files."
        (bookmarks  placeholders))
    ("sc.el"
        nil
        nil)
    ("scheme.el"
        "Scheme mode, and its idiosyncratic commands."
        (languages  lisp))
    ("score-mode.el"
        "mode for editing Gnus score files"
        (news  mail))
    ("scribe.el"
        "scribe mode, and its idiosyncratic commands."
        (wp))
    ("scroll-bar.el"
        "window system-independent scroll bar support."
        (hardware))
    ("select.el"
        "lisp portion of standard selection support."
        (internal))
    ("sendmail.el"
        "mail sending commands for Emacs."
        (mail))
    ("server.el"
        "Lisp code for GNU Emacs running as server process."
        (processes))
    ("sgml-mode.el"
        "SGML- and HTML-editing modes"
        (wp  hypermedia  comm  languages))
    ("sh-script.el"
        "shell-script editing commands for Emacs"
        (languages  unix))
    ("shadow.el"
        "Locate Emacs Lisp file shadowings."
        (lisp))
    ("shadowfile.el"
        "automatic file copying for Emacs 19"
        (comm))
    ("shell.el"
        "specialized comint.el for running the shell."
        (processes))
    ("simple.el"
        "basic editing commands for Emacs"
        nil)
    ("simula.el"
        "SIMULA 87 code editing commands for Emacs"
        (languages))
    ("site-init.el"
        nil
        nil)
    ("site-load.el"
        nil
        nil)
    ("site-start.el"
        nil
        nil)
    ("skeleton.el"
        "Lisp language extension for writing statement skeletons"
        (extensions  abbrev  languages  tools))
    ("smtpmail.el"
        nil
        (mail))
    ("solar.el"
        "calendar functions for solar events."
        (calendar))
    ("solitaire.el"
        "game of solitaire in Emacs Lisp"
        (games))
    ("sort.el"
        "commands to sort text in an Emacs buffer."
        (unix))
    ("soundex.el"
        "implement Soundex algorithm"
        (matching))
    ("spell.el"
        "spelling correction interface for Emacs."
        (wp  unix))
    ("spook.el"
        "spook phrase utility for overloading the NSA line eater"
        (games))
    ("startup.el"
        "process Emacs shell arguments"
        (internal))
    ("studly.el"
        "StudlyCaps (tm)(r)(c)(xxx)"
        (games))
    ("subr.el"
        "basic lisp subroutines for Emacs"
        nil)
    ("sun-curs.el"
        "cursor definitions for Sun windows"
        (hardware))
    ("sun-fns.el"
        "subroutines of Mouse handling for Sun windows"
        (hardware))
    ("supercite.el"
        "minor mode for citing mail and news replies"
        (mail  news))
    ("swedish.el"
        "miscellaneous functions for dealing with Swedish."
        (i18n))
    ("tabify.el"
        "tab conversion commands for Emacs"
        nil)
    ("talk.el"
        "Allow several users to talk to each other through Emacs."
        (comm  frames))
    ("tar-mode.el"
        "simple editing of tar files from GNU emacs"
        (unix))
    ("tcl-mode.el"
        "a major-mode for editing tcl/tk scripts"
        (languages  processes  tools))
    ("tcp.el"
        "TCP/IP stream emulation for GNU Emacs"
        nil)
    ("telnet.el"
        "run a telnet session from within an Emacs buffer"
        nil)
    ("tempo.el"
        "Flexible template insertion"
        (extensions  languages  tools))
    ("term-nasty.el"
        "Damned Things from terminfo.el"
        nil)
    ("term.el"
        "general command interpreter in a window stuff"
        nil)
    ("terminal.el"
        "terminal emulator for GNU Emacs."
        (comm  terminals))
    ("tex-mode.el"
        "TeX, LaTeX, and SliTeX mode commands."
        (tex))
    ("texinfmt.el"
        "format Texinfo files into Info files."
        nil)
    ("texinfo.el"
        "major mode for editing Texinfo files"
        nil)
    ("texnfo-upd.el"
        "utilities for updating nodes and menus in Texinfo files"
        (maint  tex  docs))
    ("text-mode.el"
        "text mode, and its idiosyncratic commands."
        nil)
    ("thingatpt.el"
        "Get the `thing' at point"
        (extensions  matching  mouse))
    ("time-stamp.el"
        "Maintain last change time stamps in files edited by Emacs"
        (tools))
    ("time.el"
        "display time and load in mode line of Emacs."
        nil)
    ("timer.el"
        "run a function with args at some time in future."
        nil)
    ("timezone.el"
        "time zone package for GNU Emacs"
        (news))
    ("tmm.el"
        "text mode access to menu-bar"
        nil)
    ("tpu-edt.el"
        "Emacs emulating TPU emulating EDT"
        (emulations))
    ("tpu-extras.el"
        "Scroll margins and free cursor mode for TPU-edt"
        (emulations))
    ("tpu-mapper.el"
        "Create a TPU-edt X-windows keymap file"
        (emulations))
    ("tq.el"
        "utility to maintain a transaction queue"
        (extensions))
    ("trace.el"
        "tracing facility for Emacs Lisp functions"
        (tools  lisp))
    ("two-column.el"
        "minor mode for editing of two-column text"
        nil)
    ("type-break.el"
        "encourage rests from typing at appropriate intervals"
        (extensions  timers))
    ("uncompress.el"
        "auto-decompression hook for visiting .Z files"
        nil)
    ("underline.el"
        "insert/remove underlining (done by overstriking) in Emacs."
        (wp))
    ("undigest.el"
        "digest-cracking support for the RMAIL mail reader"
        (mail))
    ("uniquify.el"
        "unique buffer names dependent on file name"
        nil)
    ("unrmail.el"
        "convert Rmail files to mailbox files."
        (mail))
    ("unused.el"
        "editing commands in GNU Emacs that turned out not to be used."
        (emulations))
    ("userlock.el"
        "handle file access contention between multiple users"
        (internal))
    ("vc-hooks.el"
        "resident support for version-control"
        nil)
    ("vc.el"
        "drive a version-control system from within Emacs"
        nil)
    ("version.el"
        "record version number of Emacs."
        (internal))
    ("vi.el"
        "major mode for emulating \"vi\" editor under GNU Emacs."
        (emulations))
    ("view.el"
        "peruse file or buffer without editing."
        nil)
    ("vip.el"
        "a VI Package for GNU Emacs"
        (emulations))
    ("viper-ex.el"
        "functions implementing the Ex commands for Viper"
        nil)
    ("viper-keym.el"
        "Viper keymaps"
        nil)
    ("viper-macs.el"
        "functions implementing keyboard macros for Viper"
        nil)
    ("viper-mous.el"
        "mouse support for Viper"
        nil)
    ("viper-util.el"
        "Utilities used by viper.el"
        nil)
    ("viper.el"
        "A full-featured Vi emulator for GNU Emacs 19 and XEmacs 19,"
        (emulations))
    ("vms-patch.el"
        "override parts of files.el for VMS."
        (vms))
    ("vms-pmail.el"
        "use Emacs as the editor within VMS mail."
        (vms))
    ("vmsproc.el"
        "run asynchronous VMS subprocesses under Emacs"
        (vms))
    ("vt-control.el"
        "Common VTxxx control functions"
        (terminals))
    ("vt100-led.el"
        "functions for LED control on VT-100 terminals & clones."
        (hardware))
    ("window.el"
        "GNU Emacs window commands aside from those written in C."
        nil)
    ("winnt.el"
        "Lisp routines for Windows NT."
        nil)
    ("ws-mode.el"
        "WordStar emulation mode for GNU Emacs"
        (emulations))
    ("x-apollo.el"
        "Apollo support functions"
        nil)
    ("x-menu.el"
        "menu support for X "
        nil)
    ("xscheme.el"
        "run Scheme under Emacs"
        (languages  lisp))
    ("xt-mouse.el"
        "Support the mouse when emacs run in an xterm."
        (mouse  terminals))
    ("yow.el"
        "quote random zippyisms"
        (games))
))

(provide 'finder-inf)

;;; finder-inf.el ends here
